/*
 * bitsdetection.c
 * check DSD file for continuous 8bit pattern
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <locale.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <asm/byteorder.h>

#include "common.h"
#include "wav.h"

#define WAVE_FORMAT_EXTENSIBLE	0xfffe

#define WAIT_TIME 5000
#define WAIT_SHORT 1000

MusicFile *musicfile = NULL;
MusicFile *p_current;
static MusicFile *p_reading;
int musicfile_current = 0;
static int musicfile_reading = 0;
int files_open;

SoundBuffer *soundbuffer = NULL;
int buffer_num;
static int soundbuffer_toread = 0;
int soundbuffer_towrite = 0;
int soundbuffer_filled = 0;


#define OPEN_RETRY_TIME 4
char devname[1024] = {0};

static int skip_sec = 0;
static int DoPcount1 = 0;
static int DoPcount2 = 0;
static SoundBuffer *psb_prev = NULL;

static const char *copyright = "Programmed & Copyright (c) 2010,2011,2012,2013,2014 by NAKATA Hiroshi <webmaster@nakata-jp.org>. All rights reserved."
#ifdef	WITHOUT_DSD
#ifdef	FFMPEG
"\nPowered by ffmpeg library (LGPL ver.2.1)."
#endif
#else
"\nCollaborated with ElectrArt"
#endif
;

const E_DSD_TYPE	e_dsd_type = E_EA;
E_DSD_TYPE	e_dsd_playback;

static int     sounddata_closefile(MusicFile *pmf)
{
    if (pmf->fp != NULL) {
    	fclose(pmf->fp);
	pmf->fp = NULL;

#ifdef	FFMPEG
        closeFfmpeg(pmf);
#endif
    }

    return (0);
}

static int     sounddata_openfile(MusicFile *pmf)
{
  pmf->fp = fopen(pmf->filename, "rb");
  if ( pmf->fp == NULL ) {
    fprintf(stderr, "file %s open error.\n", pmf->filename);
    exit (1);
  }

   if ( readWavHeader(pmf) == 0 ) {
     pmf->file_type = E_WAV;
     e_dsd_playback = E_NOTDSD;
     goto wav_opend;
   }
#ifdef	FFMPEG
   if ( readFfmpegHeader(pmf) == 0 ) {
     pmf->file_type = E_FFMPEG;
     goto wav_opend;
   }
   fprintf(stderr, "file format of %s unknown.\n", pmf->filename);
   exit (1);
#else
   rewind(pmf->fp);
   if ( readDsdHeader(pmf) == 0 ) {
     pmf->file_type = E_DSDIFF;
     goto dsd_opend;
   }

   rewind(pmf->fp);
   if ( readWsdHeader(pmf) == 0 ) {
     pmf->file_type = E_WSD;
     goto dsd_opend;
   }
   fprintf(stderr, "file %s is not WAV,DeltaSigma format.\n", pmf->filename);
   exit (1);


 dsd_opend:
   e_dsd_playback = e_dsd_type;
   pmf->total_len = (pmf->total_len + 7) / 8 * 8; // set alignment 32bit x 2
   if (e_dsd_type != E_EA) {
     pmf->wfx.Format.nSamplesPerSec *= 2;
     pmf->wfx.Format.nChannels = 2;
     if (e_dsd_type == E_DOP2) {
       if (pmf->wfx.Format.nSamplesPerSec == 176400) {
	 e_dsd_playback = E_DOP1;
       }
       while (pmf->wfx.Format.nSamplesPerSec > 176400) {
	 pmf->wfx.Format.nSamplesPerSec /= 2;
	 pmf->wfx.Format.nChannels *= 2;
       }
     }
     pmf->wfx.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
     pmf->wfx.Format.nAvgBytesPerSec = pmf->wfx.Format.nSamplesPerSec * 6;
     pmf->wfx.Format.nBlockAlign = 6;
     pmf->wfx.Format.wBitsPerSample = 24;
     pmf->wfx.Format.cbSize = 22;
     pmf->wfx.Samples.wValidBitsPerSample = pmf->wfx.Format.wBitsPerSample;
     pmf->wfx.dwChannelMask = 0;
     
     pmf->total_len += pmf->total_len / 2;	// add DoP marker length
     pmf->data_len += pmf->data_len / 2;	// add DoP marker length
   } else {
     pmf->wfx.Format.nChannels = 8;
     pmf->wfx.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
     pmf->wfx.Format.nAvgBytesPerSec = pmf->wfx.Format.nSamplesPerSec * 8;
     pmf->wfx.Format.nBlockAlign = 8;
     pmf->wfx.Format.wBitsPerSample = 8;
     pmf->wfx.Format.cbSize = 22;
     pmf->wfx.Samples.wValidBitsPerSample = pmf->wfx.Format.wBitsPerSample;
     pmf->wfx.dwChannelMask = 0;
   }
#endif	/* WITHOUT_DSD */

 wav_opend:
   pmf->sound_left_flag = TRUE;
   pmf->left_buffer_count = 0;
   pmf->fmt.p_wfx = &pmf->wfx;
#ifdef	FFMPEG
   if (pmf->file_type != E_FFMPEG) {
#endif
   switch ( pmf->wfx.Format.wBitsPerSample ) {
   case 8:
#ifdef	WITHOUT_DSD
     pmf->fmt.input_format = SND_PCM_FORMAT_U8;
#else	/* WITHOUT_DSD */
#endif	/* WITHOUT_DSD */
     break;
   case 16:
     pmf->fmt.input_format = SND_PCM_FORMAT_S16_LE;
     break;
   case 24:
     if ( pmf->wfx.Format.nBlockAlign / pmf->wfx.Format.nChannels == 4 ) {
       pmf->fmt.input_format = SND_PCM_FORMAT_S24_LE;
     } else {
       pmf->fmt.input_format = SND_PCM_FORMAT_S24_3LE;
     }
     break;
   case 32:
     if ( pmf->wfx.Format.wFormatTag == 3 ||
	  pmf->wfx.SubFormat == 3 ) {
       pmf->fmt.input_format = SND_PCM_FORMAT_FLOAT_LE;
     } else {
       pmf->fmt.input_format = SND_PCM_FORMAT_S32_LE;
     }
     break;
   default:
     exit (1);
   }
#ifdef	FFMPEG
   }
#endif
#if 0
   if ( pmf->wfx.Format.nSamplesPerSec > 384000 ) {
     fprintf(stderr, "sampling rate %dHz is too large\n", pmf->wfx.Format.nSamplesPerSec);
     exit(1);
   }
#endif
   return (0);
}

static int simple_read_audio(MusicFile *pmf)
{
  int size, result = 0;
  SoundBuffer *psb = &soundbuffer[soundbuffer_toread];
#ifndef	WITHOUT_DSD
  unsigned char *p;
  int i;
#endif

  while ( psb->buffer_start < BUFFER_SIZE ) {
    if ( pmf->data_len <= 0 ) {
      pmf->sound_left_flag = FALSE;
      if ( psb->buffer_start > 0 ) {
	psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	goto flush_buffer;      /* flush last buffer of previous sound */
      } else if ( psb_prev != NULL ) {
	psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
      }
    }
    size = BUFFER_SIZE - psb->buffer_start;
    if ( pmf->data_len < size ) {
      size = pmf->data_len;
    }
      
    unsigned char * const local_buff = (unsigned char*)psb->audio_buf + psb->buffer_start;
    switch (pmf->file_type) {
    case E_WAV:
      result = readWavData(pmf, local_buff, size);
      break;
#ifdef	FFMPEG
    case E_FFMPEG:
      while ( (result = readFfmpegData(pmf, local_buff, BUFFER_SIZE - psb->buffer_start)) == 0 )
	;	// empty loop body
      break;
#else  /* FFMPEG */
    case E_DSDIFF:
      result = readDsdData(pmf, local_buff, size, (e_dsd_type != E_EA));
      break;
    case E_WSD:
      result = readWsdData(pmf, local_buff, size, (e_dsd_type != E_EA));
      break;
#endif	/* FFMPEG */
    }
#ifndef	WITHOUT_DSD
    if (pmf->file_type != E_WAV) {
      int i;
      switch (e_dsd_playback) {
      case E_DOP1:
	for (i = 0; i < result; i += 6) {
	  local_buff[i + 2] = local_buff[i + 5] = (DoPcount2 == 0) ? 5 : 0xfa;
	  DoPcount2 = (DoPcount2 + 1) % 2;
	}
	break;
      case E_DOP2:
	for (i = 0; i < result; i += 3) {
	  local_buff[i + 2] = (DoPcount2 == 0) ? 6 : 0xf9;
	  if (++DoPcount1 >= pmf->wfx.Format.nSamplesPerSec / 88200 * 2) {
	    DoPcount1 = 0;
	    DoPcount2 = (DoPcount2 + 1) % 2;
	  }
	}
	break;
      default:
	break;
      }
    }
#endif
    if ( result <= 0 ) {
      pmf->sound_left_flag = FALSE;
      pmf->data_len = 0;
      DEBUG_PRINT("e1");
      if ( psb->buffer_start > 0 ) {
	psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	goto flush_buffer;      /* flush last buffer of previous sound */
      }
      if ( psb_prev != NULL ) {
	psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
      }
      DEBUG_PRINT("Rse");
      return (psb->buffer_start);
    }
    psb->buffer_start += result;
#ifdef	FFMPEG
    if ( pmf->file_type != E_FFMPEG ) {
      pmf->data_len -= result;
    }
#else
    pmf->data_len -= result;
#endif
    DEBUG_PRINT("r");
  }
flush_buffer:
  if ( psb_prev != NULL && psb_prev->pmf != psb->pmf ) {
    sounddata_closefile(psb_prev->pmf);
  }
  if ( psb->buffer_start > 0 ) {
    psb_prev = psb;
    CYCLIC_INCREMENT(soundbuffer_toread, buffer_num);
    memset(&soundbuffer[soundbuffer_toread], 0, sizeof(SoundBuffer));
    pmf->left_buffer_count++;
    soundbuffer_filled++;

    return (psb->buffer_start);
  }

  return (psb_prev->buffer_start);
}

static int read_audio(MusicFile *pmf)
{
  static char local_buf[1536];
  int i;
  int size, result = -1;
  int read_count = 0;
  SoundBuffer *psb = &soundbuffer[soundbuffer_toread];
  static int local_buf_start = 0;
  static int local_buf_len = 0;

  if ( soundbuffer_filled >= (buffer_num - 2) || !pmf->sound_left_flag ) {
    return (0);          /* means no error */
  }

  DEBUG_PRINT("Rs");
  memset(psb, 0, sizeof(SoundBuffer));
  psb->pmf = pmf;
  if ( psb_prev == NULL ) {
#if 0
    psb->buffer_mark |= BUFFER_MARK_AFTERCHANGE;
    psb->buffer_mark |= BUFFER_MARK_AFTERFORMATCHANGE;
#endif
  } else if ( psb_prev->pmf != pmf ) {
    psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
    psb->buffer_mark |= BUFFER_MARK_AFTERCHANGE;
    if ( psb_prev->pmf->wfx.Format.nSamplesPerSec != psb->pmf->wfx.Format.nSamplesPerSec
	 || psb_prev->pmf->wfx.Format.nBlockAlign != psb->pmf->wfx.Format.nBlockAlign
	 || psb_prev->pmf->wfx.Format.nChannels != psb->pmf->wfx.Format.nChannels ) {
      psb->buffer_mark |= BUFFER_MARK_AFTERFORMATCHANGE;
    }
  }
  const AudioSpec *pas = &(psb->pmf->fmt);
  if ( pas->output_format == pas->input_format
#ifdef	FFMPEG
 	&& pmf->file_type != E_FFMPEG
#endif
      ) {
    return (simple_read_audio(pmf));
  }

  while ( psb->buffer_start < BUFFER_SIZE ) {

    if ( local_buf_len <= 0 ) {
      if ( pmf->data_len <= 0 ) {
	pmf->sound_left_flag = FALSE;
	if ( psb->buffer_start <= 0 ) {
	  if ( psb_prev != NULL ) {
	    psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	  }
	} else {
	  psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
        }
	goto flush_buffer;      /* flush last buffer of previous sound */
      }
      size = sizeof(local_buf);
      if ( pmf->data_len < size ) {
	size = pmf->data_len;
      }
      
      switch (pmf->file_type) {
      case E_WAV:
	result = readWavData(pmf, (unsigned char *)local_buf, size);
	break;
#ifdef	FFMPEG
      case E_FFMPEG:
	while ( (result = readFfmpegData(pmf, (unsigned char *)local_buf, sizeof(local_buf))) == 0)
		;	// empty loop body
	break;
#else  /* WITHOUT_DSD */
#if 0
      case E_DSDIFF:
	result = readDsdData(pmf, (unsigned char *)local_buf, size, (e_dsd_type != E_EA));
	break;
      case E_WSD:
	result = readWsdData(pmf, (unsigned char *)local_buf, size, (e_dsd_type != E_EA));
	break;
#else
      default:
	break;
#endif
#endif	/* WITHOUT_DSD */
      }
#ifndef	WITHOUT_DSD
      if (pmf->file_type != E_WAV) {
	switch (e_dsd_playback) {
	case E_DOP1:
	  for (i = 0; i < result; i += 6) {
	    local_buf[i + 2] = local_buf[i + 5] = (DoPcount2 == 0) ? 5 : 0xfa;
	    DoPcount2 = (DoPcount2 + 1) % 2;
	  }
	  break;
	case E_DOP2:
	  for (i = 0; i < result; i += 3) {
	    local_buf[i + 2] = (DoPcount2 == 0) ? 6 : 0xf9;
	    if (++DoPcount1 >= pmf->wfx.Format.nSamplesPerSec / 88200 * 2) {
	      DoPcount1 = 0;
	      DoPcount2 = (DoPcount2 + 1) % 2;
	    }
	  }
	  break;
	default:
	  break;
	}
      }
#endif	/* WITHOUT_DSD */
      if ( result <= 0 ) {
#ifdef	WITHOUT_DSD
	if (e_dsd_playback == E_DOP2 && DoPcount1 != 0) {
	  for (i = 0; i < result; i += 3) {
	    local_buf[i] = local_buf[i + 1] = 0x96;
	    local_buf[i + 2] = (DoPcount2 == 0) ? 6 : 0xf9;
	    if (++DoPcount1 >= pmf->wfx.Format.nSamplesPerSec / 88200 * 2) {
	      DoPcount1 = 0;
	      DoPcount2 = (DoPcount2 + 1) % 2;
	    }
	  }
	}
#endif	/* WITHOUT_DSD */
	pmf->sound_left_flag = FALSE;
	pmf->data_len = 0;
	local_buf_len = 0;
	local_buf_start = 0;
	DEBUG_PRINT("e2");
	if ( psb->buffer_start > 0 ) {
	  psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	  //sounddata_closefile(psb->pmf);
	} else if ( psb_prev != NULL ) {
	  sounddata_closefile(psb_prev->pmf);
	  psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	}
	goto flush_buffer;
      }
      read_count += result;
      local_buf_len = result;
      local_buf_start = 0;
#ifdef	FFMPEG
      if ( pmf->file_type != E_FFMPEG ) {
        pmf->data_len -= result;
      }
#else
      pmf->data_len -= result;
#endif
      DEBUG_PRINT("Re");
    }

    union {
      long l;
      float f;
    } u;
    long l;
    float f;
    double d;
    unsigned char *p_src;
    unsigned char *p_dst;
    unsigned char *p_data;
    int i;

    p_dst = &(psb->audio_buf[psb->buffer_start]);
    p_src = (unsigned char *)&local_buf[local_buf_start];
    result = local_buf_len;
    size = (BUFFER_SIZE - psb->buffer_start) * pas->bytes_per_insample / pas->bytes_per_outsample;
    if ( result > size ) {
      result = size;
    }

    for ( i = 0; i < result; i += pas->bytes_per_insample ) {
      p_data = (unsigned char *)&l;
      switch ( pas->input_format ) {
      case SND_PCM_FORMAT_U8:
	l = *p_src;
	l -= 0x80;
	l <<= 24;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S16_LE:
	l = __le16_to_cpu(*(__le16 *)p_src);
	l <<= 16;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U16_LE:
	l = __le16_to_cpu(*(__le16 *)p_src);
	l <<= 16;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S16_BE:
	l = __be16_to_cpu(*(__be16 *)p_src);
	l <<= 16;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U16_BE:
	l = __be16_to_cpu(*(__be16 *)p_src);
	l <<= 16;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S32_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U32_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S32_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U32_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	l <<= 8;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	l <<= 8;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	l <<= 8;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	l <<= 8;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S18_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 6;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U18_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 6;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S18_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 6;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U18_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 6;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S20_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 4;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U20_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 4;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S20_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 4;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U20_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 4;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_FLOAT_LE:
	u.l = __le32_to_cpu(*(__le32 *)p_src);
	f = u.f;
	l = f * 0x80000000;
	d = (double)f;
	break;
      case SND_PCM_FORMAT_FLOAT_BE:
	u.l = __be32_to_cpu(*(__be32 *)p_src);
	f = u.f;
	l = f * 0x80000000;
	d = (double)f;
	break;
      case SND_PCM_FORMAT_MU_LAW:
	fprintf(stderr,
		"This program does'nt support Mu-Law format.\n");
	exit(1);
      default:
	fprintf(stderr, "Unsupported input format %d.\n",
		psb->pmf->fmt.input_format);
	exit(1);
      }
      p_src += pas->bytes_per_insample;
      f = (float)d;
	
      switch ( pas->output_format ) {
      case SND_PCM_FORMAT_S16_LE:
	p_data += 2;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S24_LE:
	l >>= 8;
	if ( l & 0x800000 ) {
	  l |= 0xff000000;
	}
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S32_LE:
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_FLOAT_LE:
	memcpy(p_dst, &f, 4);
	p_dst += 4;
	break;
      case SND_PCM_FORMAT_FLOAT64_LE:
	memcpy(p_dst, &d, 8);
	p_dst += 8;
	break;
      case SND_PCM_FORMAT_S24_3LE:
	p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S20_3LE:
	l >>= 4;
	if ( l & 0x8000000 ) {
	  l |= 0xf0000000;
	}
	p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S18_3LE:
	l >>= 6;
	if ( l & 0x4000000 ) {
	  l |= 0xfc000000;
	}
	p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S16_BE:
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	break;
      case SND_PCM_FORMAT_S24_BE:
	l >>= 8;
	if ( l & 0x800000 ) {
	  l |= 0xff000000;
	}
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_S32_BE:
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_S24_3BE:
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_S20_3BE:
	l >>= 4;
	if ( l & 0x8000000 ) {
	  l |= 0xf0000000;
	}
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_S18_3BE:
	l >>= 6;
	if ( l & 0x4000000 ) {
	  l |= 0xfc000000;
	}
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_FLOAT_BE:
	p_data = (unsigned char *)&f;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_FLOAT64_BE:
	p_data = (unsigned char *)&d;
	*p_dst++ = p_data[7];
	*p_dst++ = p_data[6];
	*p_dst++ = p_data[5];
	*p_dst++ = p_data[4];
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
	  
      case SND_PCM_FORMAT_U8:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U16_LE:
	l += 0x80000000;
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U16_BE:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	break;
      case SND_PCM_FORMAT_U24_LE:
	l += 0x80000000;
	break;
      case SND_PCM_FORMAT_U24_BE:
	l += 0x80000000;
	break;
      case SND_PCM_FORMAT_U32_LE:
	l += 0x80000000;
	*p_dst++ = p_data[0];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U32_BE:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_U24_3LE:
	l += 0x80000000;
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U24_3BE:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_U20_3LE:
	l += 0x80000000;
	l >>= 4;
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U20_3BE:
	l += 0x80000000;
	l >>= 4;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_U18_3LE:
	l += 0x80000000;
	l >>= 6;
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U18_3BE:
	l += 0x80000000;
	l >>= 6;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_IEC958_SUBFRAME_LE:
      case SND_PCM_FORMAT_IEC958_SUBFRAME_BE:
      case SND_PCM_FORMAT_MU_LAW:
      case SND_PCM_FORMAT_A_LAW:
      case SND_PCM_FORMAT_IMA_ADPCM:
      case SND_PCM_FORMAT_MPEG:
      case SND_PCM_FORMAT_GSM:
      case SND_PCM_FORMAT_SPECIAL:
      default:
	fprintf(stderr, "Unsupported output format %d \nSorry.\n",
		pas->output_format);
	exit (1);
      }
      psb->buffer_start += pas->bytes_per_outsample;
      local_buf_len -= pas->bytes_per_insample;
      local_buf_start += pas->bytes_per_insample;
    }
  }
 flush_buffer:
  if ( psb_prev != NULL && psb_prev->pmf != psb->pmf ) {
    sounddata_closefile(psb_prev->pmf);
  }
  if ( psb->buffer_start > 0 ) {
    psb_prev = psb;
    CYCLIC_INCREMENT(soundbuffer_toread, buffer_num);
    memset(&soundbuffer[soundbuffer_toread], 0, sizeof(SoundBuffer));
    pmf->left_buffer_count++;
    soundbuffer_filled++;
  }

  return (read_count);
}

static void usage(const char *command)
{
  printf("usage: %s -D device_name [-B buffsize] sound_file...\n\n", command);
}

int main(const int argc, const char **argv)
{
  char c;
  int i, j, skip, ret;
  pthread_t thread_id;
  struct sched_param sp = {50};
  void *dummy;
  int left_bit = 0, right_bit = 0;
  int last_left_bit_count = 0, last_right_bit_count = 0;
  int left_bit_count = 0, right_bit_count = 0;

  setlocale(LC_ALL, NULL);

  j = 0;
  skip = 1;
  for ( i = skip; i < argc; i++ ) {
    if ( i < argc - 1 && strncmp(argv[i], "-D", 2) == 0 ) {
      strncpy(devname, argv[i+1], 1024);
      skip += 2;
      i++;
    } else if ( strncmp(argv[i], "--version", 9) == 0 ) {
        printf(VERSION " %s\n",copyright);
	return 0;
    } else if ( i < argc - 1 && strncmp(argv[i], "--skip", 6) == 0 ) {
      if ( sscanf(argv[i + 1], "%d", &skip_sec) != 1 ) {
	skip_sec = 0;
      }
      i++;
      skip += 2;
    } else if ( i < argc - 1 && strncmp(argv[i], "-B", 2) == 0 ) {
      if ( sscanf(argv[i + 1], "%d%c", &j, &c) == 2 ) {
	switch (c) {
	case 'K':
	  j *= 1024;
	  break;
	case 'M':
	  j *= 1048576;
	  break;
	default:
	  if ( sscanf(argv[i + 1], "%d", &j) != 1 ) {
	    j = 0;
	  }
	}
      } else {
	usage(argv[0]);
	exit(1);
      }
      skip += 2;
      i++;
    } else {
      break;
    }
  }
  if (argc <= skip ) {
    /* sound filename parameter not found */
    usage(argv[0]);
    exit(1);
  }

  if ( devname[0] == '\0' ) {
    usage(argv[0]);
    exit (1);
  }

  /* set process priority */
  if (geteuid() == 0 && sched_setscheduler(0, SCHED_RR, &sp) < 0) {
    perror(NULL);
    fprintf(stderr, "sched_setscheduler() failed.\n");
    exit(1);
  }

  /* initialize semaphore */

  if ( j <= 0 ) {
    j = BUFFER_NUM * BUFFER_SIZE; /* reset buffer size */
  }
  if ( j < BUFFER_NUM * BUFFER_SIZE ) {
    fprintf(stderr, "buffer size %dBytes is too small. reset to %dBytes.\n",
	    j, BUFFER_NUM * BUFFER_SIZE);
    buffer_num = BUFFER_NUM;
  } else {
    buffer_num = j / BUFFER_SIZE;
  }
  soundbuffer = malloc(sizeof(SoundBuffer) * buffer_num);
  if ( soundbuffer == NULL ) {
    fprintf(stderr, "malloc() failed.\n");
    exit(1);
  }
  memset(soundbuffer, 0, sizeof(SoundBuffer) * buffer_num);
  files_open = argc - skip;
  musicfile = malloc(sizeof(MusicFile) * files_open);
  if ( musicfile == NULL ) {
    fprintf(stderr, "malloc() failed.\n");
    exit(1);
  }
  memset(musicfile, 0, sizeof(MusicFile) * files_open);

  p_reading = musicfile;
  /* pre open all sound file */
  for ( i = 0; i < files_open; i++ ) {
    p_reading->filename = argv[i + skip];
    p_reading->valid_flag = TRUE;
    p_reading->sound_left_flag = TRUE;
    if ( sounddata_openfile(p_reading) < 0 ) {
      exit(1);
    }
    p_reading++;
  }

  p_reading = musicfile;
  musicfile_reading = 0;
  /* pre reading to fill buffers */
  while ( musicfile_reading < files_open ) {
    while ( p_reading->sound_left_flag ) {
      if ( read_audio(p_reading) <= 0 ) {
	p_reading->sound_left_flag = FALSE;
	continue;
      }
#ifndef	WITHOUT_DSD
      printf(CR "reading file %d/%d buffer %d/%d", musicfile_reading, files_open, soundbuffer_filled, buffer_num);
      fflush(stdout);
#endif
      if ( skip_sec > p_reading->sample_output / p_reading->wfx.Format.nSamplesPerSec ) {
	p_reading->left_buffer_count--;
	CYCLIC_INCREMENT(soundbuffer_towrite, buffer_num);
	soundbuffer_filled--;
      }

      if ( soundbuffer_filled >= (buffer_num - 2) ) {
	goto buffer_full;
      }
    }
    p_reading++;
    musicfile_reading++;
  }
 buffer_full:
  printf("\r                                          \r");
  fflush(stdout);

  /* initialize sound */
  p_current = NULL;

  /* main read loop */
  while ( p_reading->sound_left_flag ) {
    if (soundbuffer_filled > 0) {
      int bit_index, byte_index;
      for (byte_index = 0; byte_index < soundbuffer[soundbuffer_towrite].buffer_start; byte_index += 2) {
        for (bit_index = 7; bit_index >= 0; bit_index--) {
          if ( ((soundbuffer[soundbuffer_towrite].audio_buf[byte_index] >> bit_index) & 1) == left_bit ) {
	    left_bit_count++;
	  } else {
	    left_bit ^= 1;
	    if (last_left_bit_count == 32 && left_bit_count == 32) {
	      printf("left_channel fake PCM\n");
	    }
	    last_left_bit_count = left_bit_count;
	    left_bit_count = 1;
	  }
          if ( ((soundbuffer[soundbuffer_towrite].audio_buf[byte_index + 1] >> bit_index) & 1) == right_bit ) {
	    right_bit_count++;
	  } else {
	    right_bit ^= 1;
	    if (last_right_bit_count == 32 && right_bit_count == 32) {
	      printf("right_channel fake PCM\n");
	    }
	    last_right_bit_count = right_bit_count;
	    right_bit_count = 1;
	  }
	}
      }
      CYCLIC_INCREMENT(soundbuffer_towrite, buffer_num);
      soundbuffer_filled--;
    }

    read_audio(p_reading);
  }

  if ( musicfile_reading > 0 ) {
    sounddata_closefile(musicfile + musicfile_reading - 1);
  }

  if ( soundbuffer != NULL ) {
    free(soundbuffer);
    soundbuffer = NULL;
  }
  if ( musicfile != NULL ) {
    free(musicfile);
    musicfile = NULL;
  }
#ifndef	WITHOUT_DSD
  cleanDsfBuffer();
#endif	/* WITHOUT_DSD */


  return 0;
}
