#ifndef _COMMON_DEFINED
#define _COMMON_DEFINED
#include <alsa/asoundlib.h>
#ifdef	WITHOUT_DSD
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#endif

#ifndef	TRUE
#define	TRUE	(-1)
#define	FALSE	0
#endif

#ifdef DEBUG
#define DEBUG_PRINT(...)	fprintf(stderr, __VA_ARGS__);fflush(stderr)
#define	DEBUG_ASSERT(...)	assert(__VA_ARGS__)
#define	CR	" "
#else
#define DEBUG_PRINT(...)	/* nothing to do */
#define DEBUG_ASSERT(...)       /* nothing to do */
#define	CR	"\r"
#endif

#define	DOP_MARK11	0x05
#define	DOP_MARK12	0xfa
#define	DOP_MARK21	0x06
#define	DOP_MARK22	0xf9
#define	DSD_ZERO_POINT	0x96

#define CYCLIC_INCREMENT(counter, divider)      \
  (counter = (counter + 1) % divider)
#define CYCLIC_PREV(counter, divider)           \
  ((counter + divider - 1) % divider)

typedef enum {E_EA, E_CQ, E_DOP_1, E_DOP_2, E_NOTDSD} E_DSD_TYPE;
#define	IS_DOP(type)	((type) != E_EA && (type) != E_CQ)

struct wfx_format {
  short wFormatTag;
  short nChannels;
  int nSamplesPerSec;
  int nAvgBytesPerSec;
  short nBlockAlign;
  short wBitsPerSample;
  short cbSize;
};
union wfx_sample {
  short wValidBitsPerSample;
  short wSamplesPerBlock;
  short wReserved;
};
struct wfx {
  struct wfx_format	Format;
  union wfx_sample	Samples;
  int dwChannelMask;
  int SubFormat;
  char data[22];
};

typedef struct AudioSpec {
  snd_pcm_format_t input_format;
  snd_pcm_format_t output_format;
  int bytes_per_insample;       /* reading data packed */
  int bytes_per_inframe;        /* frame == byte_per_sample * channels */
  int bytes_per_outsample;      /* device may needs unpacked data */
  int bytes_per_outframe;       /* frame == byte_per_outsample * channels */
  int real_channels;
  int blockSize;
  char channelMatrix[10];
  struct wfx *p_wfx;
} AudioSpec;

#ifdef	WITHOUT_DSD
#define	VERSION	"alsaplay Ver.0.9.6"
enum E_FILE_TYPE {E_FFMPEG, E_WAV};
#else	/* WITHOUT_DSD */
#define	VERSION	"udaplay Ver.0.9.6"
enum E_FILE_TYPE {E_DSDIFF, E_WSD, E_WAV};

#endif	/* WITHOUT_DSD */
typedef struct MusicFile {
  int fd;
  AudioSpec fmt;
  int sound_left_flag;
  int valid_flag;
  int left_buffer_count;
#ifdef	WITHOUT_DSD
  AVFormatContext *ic;
  AVCodecContext *icc;
  AVCodec *codec;
  AVPacket pkt;
#endif	/* WITHOUT_DSD */
  enum E_FILE_TYPE file_type;
  int64_t sample_output;
  int64_t data_len;
  int64_t total_len;
  char *filename;
  struct wfx wfx;
} MusicFile;

#define	BUFFER_NUM	64
#define BUFFER_SIZE	(1024*24)
#define BUFFER_MARK_NONE		0
#define BUFFER_MARK_BEFORECHANGE	1
#define BUFFER_MARK_AFTERCHANGE		2
#define BUFFER_MARK_AFTERFORMATCHANGE	4
typedef struct SoundBuffer {
  MusicFile *pmf;
  u_int8_t audio_buf[BUFFER_SIZE];
  int buffer_start;
  int buffer_mark;
} SoundBuffer;
#endif /* _COMMON_DEFINED */

