#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>

#define TRUE (-1)
#define FALSE 0

static int *ptrIndex = NULL;
static char *fileNamePool = NULL;
static char **fileNamePtrs = NULL;
static int ptrCount = 0;
static int ptrCountUsed = 0;
static int poolSize = 0;
static int poolSizeUsed = 0;
static int *getNameEntry()
{
  if (ptrCount == 0) {
    ptrCount = 128;
    ptrIndex = malloc(sizeof(int) * ptrCount);
  }
  return ptrIndex;  
}
static int doubleNameEntry()
{
  int *newPtr;

  if (ptrCount == 0 || ptrIndex == NULL) {
    return (getNameEntry() != NULL);
  }
  ptrCount *= 2;
  newPtr = realloc(ptrIndex, sizeof(int) * ptrCount);
  if (newPtr == NULL) {
    return FALSE;
  }
  ptrIndex = newPtr;
  return TRUE;
}
static char *getFileNamePool()
{
  if (poolSize == 0) {
    poolSize = 1024;
    fileNamePool = malloc(sizeof(char) * poolSize);
  }
  return fileNamePool;
}
static int doubleFileNamePool()
{
  char *newPtr;

  if (poolSize == 0 || fileNamePool == NULL) {
    return (getFileNamePool() != NULL);
  }
  poolSize *= 2;
  newPtr = realloc(fileNamePool, poolSize);
  if (newPtr == NULL) {
    return FALSE;
  }
  fileNamePool = newPtr;
  return TRUE;
}
static int addNameEntry(int len, const char *newName)
{
  if (ptrCountUsed == ptrCount) {
    if (!doubleNameEntry()) {
      return FALSE;
    }
  }
  while (poolSizeUsed + len + 1 > poolSize) {
    if (!doubleFileNamePool()) {
      return FALSE;
    }
  }
  ptrIndex[ptrCountUsed++] = poolSizeUsed;
  memcpy(fileNamePool + poolSizeUsed, newName, len + 1);
  poolSizeUsed += len + 1;
  return TRUE;
}

void bubbleSortFiles()
{
  int i, j;
  int tmp;

  for ( i = 0; i < ptrCountUsed; i++) {
    for ( j = 0; j < ptrCountUsed - 1; j++) {
      if (strncmp(fileNamePool + ptrIndex[j], fileNamePool + ptrIndex[j+1], 2048) > 0) {
	tmp = ptrIndex[j];
	ptrIndex[j] = ptrIndex[j+1];
	ptrIndex[j+1] = tmp;
      }
    }
  }
}

char **dir2files(int startIndex, int endIndex, const char *source[])
{
  int i;
  struct stat buf;
  struct dirent *p_entry;
  DIR *p_dir;

  if (fileNamePtrs != NULL) {
    free(fileNamePtrs);
  }
  fileNamePtrs = NULL;
  for (i = startIndex; i < endIndex; i++) {
    if (stat(source[i], &buf) < 0) {
      perror("");
      fprintf(stderr, "cannot stat %s\n", source[i]);
      continue;
    }
    if (S_ISREG(buf.st_mode)) {
      addNameEntry(strlen(source[i]), source[i]);
      continue;
    }
    if (!S_ISDIR(buf.st_mode)) {
      continue;
    }
    p_dir = opendir(source[i]);
    if (p_dir == NULL) {
      perror("");
      fprintf(stderr, "directory %s open failed.\n", source[i]);
      continue;
    }
    while ( (p_entry = readdir(p_dir)) != NULL ) {
      char fileName[2048];
      if (p_entry->d_type != DT_REG) {
	continue;
      }
      if (source[i][strlen(source[i])-1] == '/' ) {
	snprintf(fileName, sizeof(fileName) - 1, "%s%s", source[i], p_entry->d_name);
      } else {
	snprintf(fileName, sizeof(fileName) - 1, "%s/%s", source[i], p_entry->d_name);
      }
      addNameEntry(strlen(fileName), fileName);
    }
    closedir(p_dir);
  }
  bubbleSortFiles();

  fileNamePtrs = malloc(sizeof(char *) * (ptrCountUsed + 1));
  for ( i = 0; i < ptrCountUsed; i++ ) {
    fileNamePtrs[i] = fileNamePool + ptrIndex[i];
  }
  fileNamePtrs[i] = NULL;
  return fileNamePtrs;
}

void dir2filesReset()
{
  if (fileNamePtrs != NULL) {
    free(fileNamePtrs);
  }
  fileNamePtrs = NULL;
  ptrCountUsed = 0;
  poolSizeUsed = 0;
}

void dir2filesClear()
{
  if (ptrIndex != NULL) {
    free(ptrIndex);
  }
  ptrIndex = NULL;
  if (fileNamePool != NULL) {
    free(fileNamePool);
  }
  fileNamePool = NULL;
  if (fileNamePtrs != NULL) {
    free(fileNamePtrs);
  }
  fileNamePtrs = NULL;
  ptrCount = 0;
  ptrCountUsed = 0;
  poolSize = 0;
  poolSizeUsed = 0;
}
#if 0
int main(int argc, char *argv[])
{
  char **names;
  int i;

  if ((names = dir2files(1, argc, argv)) == NULL) {
    exit(1);
  }
  for ( i = 0; names[i] != NULL; i++) {
    printf("%s\n", names[i]);
  }
  dir2filesClear();
  return 0;
}
#endif
