/*
 * dsd.c
 * open DDF file.
 * Copyright (c) & Programmed 2010 by NAKATA H.
 */
#include <string.h>
#include "readwrite.h"
#include "common.h"

#define BLOCK_SIZE 4096
#define DSD_MAX_CHANNEL 6

extern int myread(int, void *, int);
extern int myseek(int fd, off_t offset, int whence);

static u_int64_t	totalLen;
static u_int64_t	chunkLen;

int readDsdData(MusicFile *pmf, unsigned char *buff, int len)
{
	unsigned char read_buff[10];
	int read_len;

	len = len / 8 * 8;
	for ( read_len = 0; read_len < len; read_len += 2 ) {
		if ( myread(pmf->fd, read_buff, (size_t)pmf->fmt.real_channels) != pmf->fmt.real_channels ) {
			while (read_len & 7) {
				buff[read_len++] = DSD_ZERO_POINT;
			}
			return (read_len);
		}
		buff[read_len] = read_buff[(int)pmf->fmt.channelMatrix[0]];
		buff[read_len + 1] = read_buff[(int)pmf->fmt.channelMatrix[1]];
	}
	return (read_len);
}

/* main file convert function	---------------------------------------------*/
int	readDsdHeader(MusicFile *pmf)
{
	int rfd = pmf->fd;
	char buff[4];
	unsigned int rate;
	unsigned short channels;

	if ( myread(rfd, buff, 4) != 4 ) {
		return (-1);
	}
	if ( memcmp(buff, "FRM8", 4) != 0 ) {
		return (-1);
	}

	if ( !read64bitBElong(rfd, &totalLen) ) {
		return (-1);
	}

	if ( myread(rfd, buff, 4) != 4 ) {
		return (-1);
	}
	if ( memcmp(buff, "DSD ", 4) != 0 ) {
		return (-1);
	}
	totalLen -= 4;

	/* check each chunks */
	while ( totalLen > 0 ) {
		/* read chunk header */
		if ( myread(rfd, buff, 4) != 4 ) {
			return (-1);
		}
		if ( !read64bitBElong(rfd, &chunkLen) ) {
			return (-1);
		}
		totalLen -= 12;
		/* branch each chunks */
		if ( memcmp(buff, "FVER", 4) == 0 ) {
			unsigned int version;
			if ( !read32bitBEint(rfd, &version) ) {
				return (-1);
			}
			if ( chunkLen != 4 ) {
				return (-1);
			}
		} else if ( memcmp(buff, "PROP", 4) == 0 ) {
			u_int64_t	propChunkLen = chunkLen;
			if ( myread(rfd, buff, 4) != 4 ) {
				return (-1);
			}
			if ( memcmp(buff, "SND ", 4) != 0 ) {
				return (-1);
			}
			propChunkLen -= 4;
			/* PROP chunk has sub chunks */
			while ( propChunkLen > 0 ) {
				u_int64_t	subChunkLen;
				if ( myread(rfd, buff, 4) != 4 ) {
					return (-1);
				}
				if ( !read64bitBElong(rfd, &subChunkLen) ) {
					return (-1);
				}
				propChunkLen -= 12;
				if ( memcmp(buff, "FS  ", 4) == 0 ) {
					if ( !read32bitBEint(rfd, &rate) ) {
						return (-1);
					}
					if ( subChunkLen != 4 ) {
						return (-1);
					}
				} else if ( memcmp(buff, "CHNL", 4) == 0 ) {
					unsigned short s;
					if ( !read16bitBEint(rfd, &channels) ) {
						return (-1);
					}
					pmf->fmt.real_channels = channels;
					if ( channels < 2 ) {
						return (-1);
					}
					if ( channels > DSD_MAX_CHANNEL ) {
						return (-1);
					}
					for ( s = 0; s < DSD_MAX_CHANNEL; s++ ) {
						pmf->fmt.channelMatrix[s] = -1;
					}
					/* set matrix to reorder channels */
					for ( s = 0; s < channels; s++ ) {
						if ( myread(rfd, buff, 4) != 4 ) {
							return (-1);
						}
						if ( memcmp(buff, "SLFT", 4) == 0 ||
						     memcmp(buff, "MLFT", 4) == 0 ) {
								 if ( pmf->fmt.channelMatrix[0] > 0 ) {
									 pmf->fmt.channelMatrix[s] = pmf->fmt.channelMatrix[0];
								 }
							pmf->fmt.channelMatrix[0] = s;
						} else if ( memcmp(buff, "SRGT", 4) == 0 ||
									memcmp(buff, "MRGT", 4) == 0 ) {
										if ( pmf->fmt.channelMatrix[1] > 0 ) {
											pmf->fmt.channelMatrix[s] = pmf->fmt.channelMatrix[1];
										}
							pmf->fmt.channelMatrix[1] = s;
						} else if ( memcmp(buff, "C   ", 4) == 0 
							|| memcmp(buff, "LFE ", 4) == 0
							|| memcmp(buff, "LS  ", 4) == 0
							|| memcmp(buff, "RS  ", 4) == 0 ) {
								pmf->fmt.channelMatrix[s] = s;
						} else {
							return (-1);
						}
					}
					for ( s = 0; s < DSD_MAX_CHANNEL && s < channels; s++ ) {
						if ( pmf->fmt.channelMatrix[s] < 0 ) {
							return (-1);
						}
					}
					if ( subChunkLen != sizeof(unsigned short) + 4 * channels ) {
						return (-1);
					}
				} else if ( memcmp(buff, "CMPR", 4) == 0 ) {
					char buff[256];
					int count, i;
					if ( myread(rfd, buff, 4) != 4 ) {
						return (-1);
					}
					if ( memcmp(buff, "DSD ", 4) != 0 &&
						memcmp(buff, "DST ", 4) != 0 ) {
						return (-1);
					}
					if ( myread(rfd, &count, 1) != 1 ) {
						return (-1);
					}
					for ( i = 0; i < count; i++ ) {
						if ( myread(rfd, buff + i, 1) != 1 ) {
							return (-1);
						}
					}
					if ( ((4 + sizeof(char) + count) % 2) == 1 ) {
						if ( myread(rfd, &i, 1) != 1 ) {
							return (-1);
						}
					}
				} else if ( memcmp(buff, "ABSS", 4) == 0 ) {
					unsigned short hours;
					unsigned int minutes, seconds, samples;

					if ( !read16bitBEint(rfd, &hours) ) {
					    return (-1);
					}
					myread(rfd, &minutes, 1);
					myread(rfd, &seconds, 1);
					if ( !read32bitBEint(rfd, &samples) ) {
						return (-1);
					}
					if ( subChunkLen != sizeof(short) + sizeof(char) * 2 + sizeof(int) ) {
						return (-1);
					}
				} else if ( memcmp(buff, "LSCO", 4) == 0 ) {
					unsigned short config;
					
					if ( !read16bitBEint(rfd, &config) ) {
						return (-1);
					}
					if ( subChunkLen != sizeof(short) ) {
						return (-1);
					}
				} else {
					myseek(rfd, subChunkLen, SEEK_CUR);
				}
				propChunkLen -= subChunkLen;
				if ( (subChunkLen % 2) == 1 ) {
					propChunkLen--;
				}
			}
		} else if ( memcmp(buff, "DSD ", 4) == 0 ) {
			if ( channels == 2 ) {
				pmf->data_len = pmf->total_len = chunkLen;
			} else {
				pmf->data_len = pmf->total_len = chunkLen * 2 / channels;
			}
			pmf->wfx.Format.nSamplesPerSec = (rate / 32);
			return (0);
		} else if ( memcmp(buff, "DST ", 4) == 0 ) {
			return (-1);
	    } else if ( memcmp(buff, "DSTI", 4) == 0 ) {
			return (-1);
		} else if ( memcmp(buff, "COMT", 4) == 0 ) {
			myseek(rfd, (int)chunkLen, SEEK_CUR);
		} else if ( memcmp(buff, "DIIN", 4) == 0 ) {
			myseek(rfd, (int)chunkLen, SEEK_CUR);
		} else if ( memcmp(buff, "MANF", 4) == 0 ) {
			myseek(rfd, (int)chunkLen, SEEK_CUR);
		} else {
			myseek(rfd, (int)chunkLen, SEEK_CUR);
		}
		totalLen -= chunkLen;
		if ( (chunkLen % 2) == 1 ) {
			myseek(rfd, 1, SEEK_CUR);
			totalLen--;
		}
	}

	return (-1);
}
/* end of file	=============================================================*/
