#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#define	READ_LEN	512
static unsigned char remember[READ_LEN];
static int remember_offset = 0;
static int remember_len = 0;
int myread(int fd, void *buf, int len)
{
	unsigned char *p = buf;
	int size;
	int ret;
	int readlen = 0;

	if (remember_len > 0) {
		if (remember_len >= len) {
			memcpy(buf, remember + remember_offset, len);
			remember_offset += len;
			remember_len -= len;
			return len;
		}
		memcpy(p, remember + remember_offset, remember_len);
		readlen = remember_len;
		len -= remember_len;
		p += remember_len;
	}
	remember_offset = 0;
	remember_len = 0;
	while (len >= READ_LEN) {
		ret = read(fd, p, READ_LEN);
		usleep(0);		// give chance of thread switch
		if (ret < 0) {
			return ret;	// error happen
		}
		if (ret == 0) {
			return readlen;	// may be EOF
		}
		readlen += ret;
		len -= ret;
		p += ret;
		if (ret != READ_LEN) {
			return readlen;	// may be EOF
		}
	}
	if (len == 0) {
		return readlen;		// successfully end
	}
	ret = read(fd, remember, READ_LEN);
	if (ret < 0) {
		return ret;		// error happen
	}
	if (ret == 0) {
		return readlen;		// may be EOF
	}
	size = ret;
	if (len < size) {
		size = len;
	}
	memcpy(p, remember, size);
	readlen += size;
	remember_offset = size;
	remember_len = ret - size;

	return readlen;
}

int myseek(int fd, off_t offset, int whence)
{
	off_t rest = offset % READ_LEN;
	off_t pos = offset - rest;
	int ret;

	switch (whence) {
	case SEEK_SET:
		lseek(fd, pos, SEEK_SET);
		ret = read(fd, remember, READ_LEN);
		if (ret < 0) {
			return -1;
		}
		if (ret < rest) {
			return -1;
		}
		remember_len = ret - rest;
		remember_offset = rest;
		return 0;
	case SEEK_CUR:
		if (offset > 0 ) {
			if (offset <= remember_len) {
				remember_len -= offset;
				remember_offset += offset;
				return 0;
			}
			offset -= remember_len;
			rest = offset % READ_LEN;
			pos = offset - rest;

			if (pos > 0) {
				lseek(fd, pos, SEEK_CUR);
			}
			ret = read(fd, remember, READ_LEN);
			if (ret < 0) {
				return -1;
			}
			if (ret < rest) {
				return -1;
			}
			remember_len = ret - rest;
			remember_offset = rest;
			return 0;
		}
		break;
	default:
		break;
	}
	return -1;
}
